//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#include "UPV.h"

ClassImp(UPV)

////////////////////////////////////////////////////////////////////////////////////
UPVsource::UPVsource(void){ 
////////////////////////////////////////////////////////////////////////////////////
  source_files    = "none";
  source_name     = "NONE";
  outdir          = ".";
  source_nc       = 0;
  coinc_livetime  = 0.0;
  source_livetime = 0.0;
  source_coinc_nc = 0;
  target_nc       = 0;
  target_coinc_nc = 0;
  status          = true;
  veto_dt         = 0.0;
  veto_nseg       = 0;
  source_veto_nc  = 0;
  source_veto_u   = 0;
  target_veto_u   = 0;
}

////////////////////////////////////////////////////////////////////////////////////
UPVsource::~UPVsource(void){ 
////////////////////////////////////////////////////////////////////////////////////
}


////////////////////////////////////////////////////////////////////////////////////
UPV::UPV(const string aOptFile, Segments *aInSeg){ 
////////////////////////////////////////////////////////////////////////////////////

  // timer starts
  time ( &timer_start );

  // analysis segments
  insegments=aInSeg;
  cursegments = new Segments(insegments->GetStarts(),insegments->GetEnds());
  status_OK=cursegments->GetStatus();
  
  // read option files
  status_OK*=GetParameters(aOptFile,cursegments);
  optionfile=aOptFile;
  subtitle="";
  
  // read target triggers
  target = new TriggerPlot((int)snrthr.size(),target_trigfiles,"","STANDARD",verbosity);
  status_OK*=target->GetStatus();

  // clusterize target triggers
  target->SetClusterizeDt(target_clusterdt);
  target->SetClusterizeSNRthr(target_snrmin);
  if(!target->Clusterize(1)){
    cerr<<"UPV::UPV: clustering of target triggers failed"<<endl;
    status_OK=false;
  }

  // create output directory
  system(("mkdir -p "+outdir+"/"+target->GetName()+"/perf").c_str());

  // init source
  source_index=-1;
  source=NULL;

  // source results objects
  source_results = new UPVsource * [GetNsource()];
  for(int s=0; s<GetNsource(); s++) source_results[s] = new UPVsource();

  // coinc
  C2 = new Coinc2(verbosity);
  C2->SetCoincDt(C2_dt);
  if(target_freqmax<target->GetFrequencyMax()) C2->SetFrequencyMax(1,target_freqmax);
  if(target_freqmin>target->GetFrequencyMin()) C2->SetFrequencyMin(1,target_freqmin);

  // coinc monitors (will be updated)
  snrsnr         = new TGraph(); snrsnr->SetName("snrsnr");

  // source monitors (will be updated)
  source_c       = new TH2D("source_c",      "clusters",                  1,1,4096,1,1,100);
  source_c_cum   = new TH2D("source_c_cum",  "clusters (cumulative)",     1,1,4096,1,1,100);
  source_ct      = new TH2D("source_ct",     "cluster time",              1,1,4096,1,1,100);
  source_ct_cum  = new TH2D("source_ct_cum", "cluster time (cumulative)", 1,1,4096,1,1,100);
  source_co      = new TH2D("source_co",     "coincs",                    1,1,4096,1,1,100);
  source_co_cum  = new TH2D("source_co_cum", "coincs (cumulative)",       1,1,4096,1,1,100);
  target_co      = new TH2D("target_co",     "coincs",                    1,1,4096,1,1,100);
  target_co_cum  = new TH2D("target_co_cum", "coincs (cumulative)",       1,1,4096,1,1,100);
  source_up      = new TH2D("source_up",     "use-percentage",            1,1,4096,1,1,100);
  source_up_cum  = new TH2D("source_up_cum", "cumulative use-percentage", 1,1,4096,1,1,100);
 
  // target monitors (will not be updated)
  TriggerSelect *tselect = new TriggerSelect(target);
  tselect->SetFrequencyResolution(TMath::Min(100,target->GetNclusters()/1000+3));
  if(target_freqmin>target->GetFrequencyMin())  tselect->SetFrequencyMin(target_freqmin);
  if(target_freqmax<target->GetFrequencyMax())  tselect->SetFrequencyMax(target_freqmax);
  tselect->SetSNRRange(target_snrmin,TMath::Max(100.0,2*tselect->GetSNRMin()));
  tselect->SetSNRResolution(60);
  target_c = tselect->GetTH2D("Frequency","SNR");
  target_c->SetName(("target_c_"+target->GetName()).c_str());
  target_c->SetTitle((target->GetName()+" clusters").c_str());
  target_c->GetXaxis()->SetTitle("Frequency [Hz]");
  target_c->GetYaxis()->SetTitle("SNR");
  target_c->GetZaxis()->SetTitle("clusters");
  target_c_cum = tselect->GetTH2D("Frequency","SNR");
  target_c_cum->SetName(("target_c_cum_"+target->GetName()).c_str());
  target_c_cum->SetTitle((target->GetName()+" clusters (cumulative)").c_str());
  target_c_cum->GetXaxis()->SetTitle("Frequency [Hz]");
  target_c_cum->GetYaxis()->SetTitle("SNR threshold");
  target_c_cum->GetZaxis()->SetTitle("clusters");
  delete tselect;

  // veto threshold (will be updated)
  veto_thr = new TH1D("veto_thr", "veto_thr", 1,1,4096);

  // configure veto perf
  vp = new VetoPerf(snrthr);
  status_OK*=vp->SetOutputDirectory(outdir+"/"+target->GetName()+"/perf");
  if(target_freqmin>target->GetFrequencyMin()) vp->SetTriggerMinimumFrequency(target_freqmin);
  if(target_freqmax<target->GetFrequencyMax()) vp->SetTriggerMaximumFrequency(target_freqmax);
  vp->SetTriggers(target,cursegments);

}

////////////////////////////////////////////////////////////////////////////////////
UPV::~UPV(void){
////////////////////////////////////////////////////////////////////////////////////
  if(verbosity>1) cout<<"UPV::~UPV"<<endl;
  delete cursegments;
  delete target;
  if(source!=NULL) delete source;
  delete C2;
  for(int s=0; s<GetNsource(); s++) delete source_results[s];
  delete source_results;
  delete snrsnr;
  delete source_c;     delete source_c_cum;
  delete source_ct;    delete source_ct_cum;
  delete source_co;    delete source_co_cum;
  delete target_co;    delete target_co_cum;
  delete source_up;    delete source_up_cum;
  delete target_c;     delete target_c_cum;
  delete veto_thr;
  delete vp;
}

////////////////////////////////////////////////////////////////////////////////////
bool UPV::LoadSourceTriggers(const int aSourceIndex){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK){
    cerr<<"UPV::LoadSourceTriggers: the UPV object is corrupted"<<endl;
    return false;
  }
  if(aSourceIndex<0||aSourceIndex>=GetNsource()){
    cerr<<"UPV::LoadSourceTriggers: the source index "<<aSourceIndex<<" is not possible"<<endl;
    return false;
  }

  // new source index
  source_index=aSourceIndex;
    
  // load source triggers
  if(source!=NULL) delete source;
  source = new ReadTriggers(source_trigfiles[source_index],"",verbosity);

  // check for sanity
  if(!source->GetStatus()){
    cerr<<"UPV::LoadSourceTriggers: source triggers are corrupted ("<<source->GetName()<<")"<<endl;
    delete source; source=NULL; source_index = -1;
    return false;
  }

  // clusterize
  source->SetClusterizeDt(source_clusterdt);
  if(!source->Clusterize(1)){
    cerr<<"UPV::LoadSourceTriggers: clustering of source triggers failed ("<<source->GetName()<<")"<<endl;
    delete source; source=NULL; source_index = -1;
    return false;
  }

  // save results for this source channel
  delete source_results[source_index];
  source_results[source_index] = new UPVsource();
  source_results[source_index]->source_files = source_trigfiles[source_index];
  source_results[source_index]->source_name = source->GetName();
  source_results[source_index]->outdir = outdir+"/"+target->GetName()+"/"+source->GetName();

  // source livetime
  Segments *tmp = new Segments(source->GetSegments()->GetStarts(),source->GetSegments()->GetEnds());
  tmp->Intersect(cursegments);
  source_results[source_index]->source_livetime=tmp->GetLiveTime();
  delete tmp;

  // create output directory
  system(("mkdir -p "+source_results[source_index]->outdir).c_str());

  // init coinc
  // note: full reset: *all* clusters are considered.
  //       source/target clusters are rejected based on Coinc settings and cursegments
  if(!C2->SetTriggerSamples(source, target, 0.0, 0.0, cursegments,false)){
    cerr<<"UPV::LoadSourceTriggers: set source/target trigger samples failed"<<endl;
    delete source; source=NULL; source_index = -1;
    return false;
  }

  // update results
  source_results[source_index]->source_nc=C2->GetNActiveClusters(0);
  source_results[source_index]->target_nc=C2->GetNActiveClusters(1);
  source_results[source_index]->coinc_livetime=C2->GetCoincLiveTime();
  
  // reset monitors for this source
  ResetMonitors();
  
  return true;
}

////////////////////////////////////////////////////////////////////////////////////
bool UPV::MakeTriggerMonitors(const int aTriggerIndex){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK){
    cerr<<"UPV::MakeTriggerMonitors: the UPV object is corrupted"<<endl;
    return false;
  }
  if(source==NULL){
    cerr<<"UPV::MakeTriggerMonitors: the source triggers must be defined first"<<endl;
    return false;
  }

  // link to source/target objects
  ReadTriggers *RT;
  TH2D *CH, *CH_cum, *CTH, *CTH_cum;
  if(aTriggerIndex==0){      RT=source; CH=source_c; CH_cum=source_c_cum; CTH=source_ct; CTH_cum=source_ct_cum; }
  else if(aTriggerIndex==1){ RT=target; CH=target_c; CH_cum=target_c_cum;}
  else{
    cerr<<"UPV::MakeTriggerMonitors: the trigger index is not possible"<<endl;
    return false;
  }

  // fill monitors
  for(int c=0; c<RT->GetNclusters(); c++){// loop over clusters
    if(RT->GetClusterTag(c)<0) continue; // non-valid cluster
    CH->Fill(RT->GetClusterFrequency(c),RT->GetClusterSNR(c));
    if(!aTriggerIndex) CTH->Fill(RT->GetClusterFrequency(c),RT->GetClusterSNR(c), TMath::Min(RT->GetClusterTime(c)+C2->GetTimeOffset(0)+C2->GetCoincDt()/2.0,RT->GetClusterTimeEnd(c)+C2->GetTimeOffset(0))-TMath::Max(RT->GetClusterTime(c)+C2->GetTimeOffset(0)-C2->GetCoincDt()/2.0,RT->GetClusterTimeStart(c)+C2->GetTimeOffset(0)));
  }

  // save number of active clusters
  if(aTriggerIndex) source_results[source_index]->target_nc = CH->GetEntries();
  else              source_results[source_index]->source_nc = CH->GetEntries();

  // cumulative monitors
  for(int f=0; f<=CH->GetNbinsX()+1; f++){
    CH_cum->SetBinContent(f,CH->GetNbinsY()+1,CH->GetBinContent(f,CH->GetNbinsY()+1));
    if(!aTriggerIndex) CTH_cum->SetBinContent(f,CH->GetNbinsY()+1,CTH->GetBinContent(f,CH->GetNbinsY()+1));
    for(int s=CH->GetNbinsY(); s>=0; s--){
      CH_cum->SetBinContent(f,s,CH_cum->GetBinContent(f,s+1)+CH->GetBinContent(f,s));
      if(!aTriggerIndex) CTH_cum->SetBinContent(f,s,CTH_cum->GetBinContent(f,s+1)+CTH->GetBinContent(f,s));
    }
  }

  return true;
}

////////////////////////////////////////////////////////////////////////////////////
bool UPV::MakeCoincMonitors(void){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK){
    cerr<<"UPV::MakeCoincMonitors: the UPV object is corrupted"<<endl;
    return false;
  }
  if(source==NULL){
    cerr<<"UPV::MakeCoincMonitors: the source triggers must be defined first"<<endl;
    return false;
  }

  // init
  int prev=-1;

  // loop over coinc events
  for(int co=0; co<C2->GetNCoinc(); co++){// loop over coincs (all coincs are valid)
    snrsnr->SetPoint(co, source->GetClusterSNR(C2->GetClusterIndex(0,co)), target->GetClusterSNR(C2->GetClusterIndex(1,co)));
    target_co->Fill(source->GetClusterFrequency(C2->GetClusterIndex(0,co)), source->GetClusterSNR(C2->GetClusterIndex(0,co)));
    if(C2->GetClusterIndex(0,co)==prev) continue; // remove double counting
    source_co->Fill(source->GetClusterFrequency(C2->GetClusterIndex(0,co)),source->GetClusterSNR(C2->GetClusterIndex(0,co)));
    prev=C2->GetClusterIndex(0,co);
  }

  // save number of clusters involved in a coinc
  source_results[source_index]->source_coinc_nc = source_co->GetEntries();
  source_results[source_index]->target_coinc_nc = target_co->GetEntries();

  // cumulative monitors
  for(int f=0; f<=source_co->GetNbinsX()+1; f++){
    source_co_cum->SetBinContent(f,source_co->GetNbinsY()+1,source_co->GetBinContent(f,source_co->GetNbinsY()+1));
    target_co_cum->SetBinContent(f,source_co->GetNbinsY()+1,target_co->GetBinContent(f,source_co->GetNbinsY()+1));
    for(int s=source_co->GetNbinsY(); s>=0; s--){
      source_co_cum->SetBinContent(f,s,source_co_cum->GetBinContent(f,s+1)+source_co->GetBinContent(f,s));
      target_co_cum->SetBinContent(f,s,target_co_cum->GetBinContent(f,s+1)+target_co->GetBinContent(f,s));
    }
  }

  // axis labels
  snrsnr->GetHistogram()->SetXTitle((source->GetName()+" SNR").c_str());
  snrsnr->GetHistogram()->SetYTitle((target->GetName()+" SNR").c_str());

  return true;
}

////////////////////////////////////////////////////////////////////////////////////
void UPV::MakeProcMonitors(void){
////////////////////////////////////////////////////////////////////////////////////
  // proc monitors
  source_up->Divide(source_co,source_c);
  source_up_cum->Divide(source_co_cum,source_c_cum);
  source_up->GetZaxis()->SetRangeUser(0.0,1.0);
  source_up_cum->GetZaxis()->SetRangeUser(0.0,1.0);

  return;
}

////////////////////////////////////////////////////////////////////////////////////
void UPV::ResetMonitors(void){
////////////////////////////////////////////////////////////////////////////////////

  // ******************** COINC
  delete snrsnr;
  snrsnr = new TGraph();
  snrsnr->SetName("snrsnr");
  snrsnr->SetTitle("Coinc SNR vs. SNR");
  snrsnr->GetHistogram()->SetXTitle((source->GetName()+" SNR").c_str());
  snrsnr->GetHistogram()->SetYTitle((target->GetName()+" SNR").c_str());
  snrsnr->SetMarkerStyle(6);

  // ******************** SOURCE
  TriggerSelect *tselect = new TriggerSelect(source);
  tselect->SetFrequencyResolution(TMath::Min(100,source->GetNClusters()/1000+3));
  tselect->SetSNRRange(tselect->GetSNRMin(),TMath::Max(100.0,2*tselect->GetSNRMin()));
  tselect->SetSNRResolution(60);

  // clusters
  delete source_c; source_c = tselect->GetTH2D("Frequency","SNR");
  source_c->SetName(("source_c_"+source->GetName()).c_str());
  source_c->SetTitle((source->GetName()+" clusters").c_str());
  source_c->GetXaxis()->SetTitle("Frequency [Hz]");
  source_c->GetYaxis()->SetTitle("SNR");
  source_c->GetZaxis()->SetTitle("clusters");
  delete source_c_cum; source_c_cum = tselect->GetTH2D("Frequency","SNR");
  source_c_cum->SetName(("source_c_cum_"+source->GetName()).c_str());
  source_c_cum->SetTitle((source->GetName()+" clusters (cumulative)").c_str());
  source_c_cum->GetXaxis()->SetTitle("Frequency [Hz]");
  source_c_cum->GetYaxis()->SetTitle("SNR threshold");
  source_c_cum->GetZaxis()->SetTitle("clusters");
  
  // cluster time
  delete source_ct; source_ct = tselect->GetTH2D("Frequency","SNR");
  source_ct->SetName(("source_ct_"+source->GetName()).c_str());
  source_ct->SetTitle((source->GetName()+" cluster time").c_str());
  source_ct->GetXaxis()->SetTitle("Frequency [Hz]");
  source_ct->GetYaxis()->SetTitle("SNR");
  source_ct->GetZaxis()->SetTitle("cluster time [s]");
  delete source_ct_cum; source_ct_cum = tselect->GetTH2D("Frequency","SNR");
  source_ct_cum->SetName(("source_ct_cum_"+source->GetName()).c_str());
  source_ct_cum->SetTitle((source->GetName()+" cluster time (cumulative)").c_str());
  source_ct_cum->GetXaxis()->SetTitle("Frequency [Hz]");
  source_ct_cum->GetYaxis()->SetTitle("SNR threshold");
  source_ct_cum->GetZaxis()->SetTitle("cluster time [s]");
  
  // coincs (source)
  delete source_co; source_co = tselect->GetTH2D("Frequency","SNR");
  source_co->SetName(("source_co_"+source->GetName()).c_str());
  source_co->SetTitle((source->GetName()+": coincs").c_str());
  source_co->GetXaxis()->SetTitle("Frequency [Hz]");
  source_co->GetYaxis()->SetTitle("SNR");
  source_co->GetZaxis()->SetTitle("source clusters in a coinc");
  delete source_co_cum; source_co_cum = tselect->GetTH2D("Frequency","SNR");
  source_co_cum->SetName(("source_co_cum_"+source->GetName()).c_str());
  source_co_cum->SetTitle((source->GetName()+": coincs (cumulative)").c_str());
  source_co_cum->GetXaxis()->SetTitle("Frequency [Hz]");
  source_co_cum->GetYaxis()->SetTitle("SNR threshold");
  source_co_cum->GetZaxis()->SetTitle("source clusters in a coinc");

  // coincs (target)
  delete target_co; target_co = tselect->GetTH2D("Frequency","SNR");
  target_co->SetName(("target_co_"+source->GetName()).c_str());
  target_co->SetTitle((target->GetName()+": coincs (source ref)").c_str());
  target_co->GetXaxis()->SetTitle("Frequency [Hz]");
  target_co->GetYaxis()->SetTitle("SNR");
  target_co->GetZaxis()->SetTitle("target clusters in a coinc");
  delete target_co_cum; target_co_cum = tselect->GetTH2D("Frequency","SNR");
  target_co_cum->SetName(("target_co_cum_"+source->GetName()).c_str());
  target_co_cum->SetTitle((target->GetName()+": coincs (cumulative) (source ref)").c_str());
  target_co_cum->GetXaxis()->SetTitle("Frequency [Hz]");
  target_co_cum->GetYaxis()->SetTitle("SNR threshold");
  target_co_cum->GetZaxis()->SetTitle("target clusters in a coinc");

  // use-percentage
  delete source_up; source_up = tselect->GetTH2D("Frequency","SNR");
  source_up->SetName(("source_up_"+source->GetName()).c_str());
  source_up->SetTitle((source->GetName()+": use-percentage").c_str());
  source_up->GetXaxis()->SetTitle("Frequency [Hz]");
  source_up->GetYaxis()->SetTitle("SNR");
  source_up->GetZaxis()->SetTitle("use-percentage");
  source_up->GetZaxis()->SetRangeUser(0.0,1.0);
  delete source_up_cum; source_up_cum = tselect->GetTH2D("Frequency","SNR");
  source_up_cum->SetName(("source_up_cum_"+source->GetName()).c_str());
  source_up_cum->SetTitle((source->GetName()+": use-percentage (cumulative)").c_str());
  source_up_cum->GetXaxis()->SetTitle("Frequency [Hz]");
  source_up_cum->GetYaxis()->SetTitle("SNR threshold");
  source_up_cum->GetZaxis()->SetTitle("use-percentage");
  source_up_cum->GetZaxis()->SetRangeUser(0.0,1.0);

  // ******************** TARGET
  target_c->Reset();
  target_c_cum->Reset();

  //******************** VETO
  delete veto_thr; veto_thr = tselect->GetTH1D("Frequency");
  veto_thr->SetName(("veto_thr_"+source->GetName()).c_str());
  veto_thr->SetTitle((source->GetName()+": veto SNR threshold").c_str());
  veto_thr->GetXaxis()->SetTitle("Frequency [Hz]");
  veto_thr->GetYaxis()->SetTitle("SNR threshold");
  veto_thr->SetLineColor(1);
  veto_thr->SetLineWidth(3);
  for(int f=0; f<=veto_thr->GetNbinsX()+1; f++) veto_thr->SetBinContent(f,tselect->GetSNRMax());

  return;  
}

////////////////////////////////////////////////////////////////////////////////////
double UPV::MakeVetoThreshold(void){
////////////////////////////////////////////////////////////////////////////////////

  // init
  source_results[source_index]->source_veto_nc=0;
  source_results[source_index]->veto_dt=0.0;
  source_results[source_index]->source_veto_u=0;
  source_results[source_index]->target_veto_u=0;

  // SNR threshold = f(frequency)
  int s;
  for(int f=0; f<=source_up_cum->GetNbinsX()+1; f++){
    for(s=0; s<=source_up_cum->GetNbinsY()+1; s++){
      veto_thr->SetBinContent(f,source_up_cum->GetYaxis()->GetBinLowEdge(s));
      if(source_up_cum->GetBinContent(f,s) > veto_upmin && source_co_cum->GetBinContent(f,s) > veto_umin) break;// above thresholds
    }
    if(s>=source_up_cum->GetNbinsY()+1) continue; // infinite threshold 
    source_results[source_index]->source_veto_nc+=source_c_cum->GetBinContent(f,s);
    source_results[source_index]->veto_dt+=source_ct_cum->GetBinContent(f,s);
    source_results[source_index]->source_veto_u+=source_co_cum->GetBinContent(f,s);
    source_results[source_index]->target_veto_u+=target_co_cum->GetBinContent(f,s);
  }

  return source_results[source_index]->veto_dt;
}

////////////////////////////////////////////////////////////////////////////////////
Segments* UPV::GetVetoSegments(void){
////////////////////////////////////////////////////////////////////////////////////
  if(source==NULL){
    cerr<<"UPV::GetVetoSegments: the source trigger sample is not defined"<<endl;
    return NULL;
  }

  // make veto segments
  Segments *vetoseg = new Segments();  
  int binn;
  for(int c0=0; c0<source->GetNclusters(); c0++){
    if(source->GetClusterTag(c0)==-1) continue;
    binn = veto_thr->FindFixBin(source->GetClusterFrequency(c0));
    if(veto_thr->GetBinContent(binn)>=source_up->GetYaxis()->GetBinLowEdge(source_up->GetNbinsY()+1)) continue;// infinite threshold
    if(source->GetClusterSNR(c0)<=veto_thr->GetBinContent(binn)) continue; // below SNR threshold
    vetoseg->AddSegment(TMath::Max(source->GetClusterTime(c0)+C2->GetTimeOffset(0)-C2->GetCoincDt()/2.0,source->GetClusterTimeStart(c0)+C2->GetTimeOffset(0)), TMath::Min(source->GetClusterTime(c0)+C2->GetTimeOffset(0)+C2->GetCoincDt()/2.0,source->GetClusterTimeEnd(c0)+C2->GetTimeOffset(0)));
  }

  // number of veto segments
  source_results[source_index]->veto_nseg = vetoseg->GetNsegments();

  return vetoseg;
}

////////////////////////////////////////////////////////////////////////////////////
bool UPV::RemoveVetoedTriggers(void){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK){
    cerr<<"UPV::RemoveVetoedSegments: the UPV object is corrupted"<<endl;
    return false;
  }
  if(source==NULL){
    cerr<<"UPV::RemoveVetoedSegments: the source trigger sample is not defined"<<endl;
    return false;
  }

  // make veto segments
  Segments *veto_coincseg = new Segments();
  int binn;
  for(int co=0; co<C2->GetNCoinc(); co++){// loop over coincs (all coincs are valid)
    
    // coinc is selected for veto?
    binn = veto_thr->FindFixBin(source->GetClusterFrequency(C2->GetClusterIndex(0,co)));
    if(veto_thr->GetBinContent(binn)>=source_up->GetYaxis()->GetBinLowEdge(source_up->GetNbinsY()+1)) continue;// infinite threshold
    if(source->GetClusterSNR(C2->GetClusterIndex(0,co))<=veto_thr->GetBinContent(binn)) continue; // below SNR threshold
    
    // segment to remove:
    veto_coincseg->AddSegment(TMath::Max(source->GetClusterTime(C2->GetClusterIndex(0,co))+C2->GetTimeOffset(0)-C2->GetCoincDt()/2.0,source->GetClusterTimeStart(C2->GetClusterIndex(0,co))+C2->GetTimeOffset(0)), TMath::Min(source->GetClusterTime(C2->GetClusterIndex(0,co))+C2->GetTimeOffset(0)+C2->GetCoincDt()/2.0,source->GetClusterTimeEnd(C2->GetClusterIndex(0,co))+C2->GetTimeOffset(0)));
  }

  // remove vetoed segments
  bool res=veto_coincseg->Reverse()&&cursegments->Intersect(veto_coincseg);

  delete veto_coincseg;
  return res;
}

////////////////////////////////////////////////////////////////////////////////////
void UPV::PrintMonitors(void){
////////////////////////////////////////////////////////////////////////////////////
  C2->SetGridx(1);
  C2->SetGridy(1);

  C2->SetLogx(1); C2->SetLogy(1); C2->SetLogz(1);
  C2->Draw(source_c_cum,"colz");
  C2->Draw(veto_thr,"same");
  C2->Print(source_results[source_index]->outdir+"/source_c_cum.png");
  C2->Draw(target_c_cum,"colz");
  C2->Print(source_results[source_index]->outdir+"/target_c_cum.png");
  C2->Draw(source_co_cum,"colz");
  C2->Draw(veto_thr,"same");
  C2->Print(source_results[source_index]->outdir+"/source_co_cum.png");
  if(snrsnr->GetN()){
    C2->Draw(snrsnr,"ap");
    C2->Print(source_results[source_index]->outdir+"/coinc_snrsnr.png");
  }
  
  C2->SetLogx(1); C2->SetLogy(1); C2->SetLogz(0);
  C2->Draw(source_up_cum,"colz");
  C2->Draw(veto_thr,"same");
  C2->Print(source_results[source_index]->outdir+"/source_up_cum.png");
    
  return;
}

////////////////////////////////////////////////////////////////////////////////////
void UPV::PrintVeto(void){
////////////////////////////////////////////////////////////////////////////////////

  // veto segment list
  Segments *vseg = GetVetoSegments();
  if(vseg==NULL) return;
  string vfile=source_results[source_index]->outdir+"/veto_"+source_results[source_index]->source_name+".txt";
  vseg->Write(vfile);
  delete vseg;

  // veto recipe
  TFile *frecipe = new TFile((source_results[source_index]->outdir+"/veto_recipe_"+source_results[source_index]->source_name+".root").c_str(),"RECREATE");
  frecipe->cd();
  veto_thr->Write();
  frecipe->Close();
  return;
}

////////////////////////////////////////////////////////////////////////////////////
bool UPV::MakeHtml(vector <int> aIndexes){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK){
    cerr<<"UPV::MakeHtml: the UPV object is corrupted"<<endl;
    return false;
  }

  // import material
  system(("cp -f ${UPV_HTML}/style.css "+outdir+"/style.css").c_str());
  system(("cp -f ${UPV_HTML}/qm.gif "+outdir+"/qm.gif").c_str());
  system(("cp -f ${GWOLLUM_DOC}/Pics/gwollum_logo_min_trans.gif "+outdir+"/icon.gif").c_str());

  // index header & scripts
  ofstream report((outdir+"/index.html").c_str());
  report<<"<html>"<<endl;
  report<<"<head>"<<endl;
  report<<"<title>UPV Report</title>"<<endl;
  report<<"<link rel=\"stylesheet\" href=\"style.css\" type=\"text/css\" />"<<endl;
  report<<"<link rel=\"icon\" type=\"image/x-icon\" href=\"./icon.gif\" />"<<endl;
  report<<"<link rel=\"shortcut icon\" type=\"image/x-icon\" href=\"./icon.gif\" />"<<endl;
  report<<"<script type=\"text/javascript\">"<<endl;
  report<<"function showPlot(target, source, type) {"<<endl;
  report<<"  var basename ="<<endl;
  report<<"    source + \"_\" + type;"<<endl;
  report<<"    document.getElementById(\"plot_\" + source).src ="<<endl;
  report<<"    \"./\" + target + \"/\" + source + \"/\" + type +\".png\";"<<endl;
  report<<"}"<<endl;
  report<<"</script>"<<endl;
  report<<"<script type=\"text/javascript\">"<<endl;
  report<<"function toggle(anId) {"<<endl;
  report<<"  node = document.getElementById(anId);"<<endl;
  report<<"  if (node.style.visibility==\"hidden\") {"<<endl;
  report<<"    node.style.visibility = \"visible\";"<<endl;
  report<<"    node.style.height = \"auto\";"<<endl;
  report<<"  }"<<endl;
  report<<"  else {"<<endl;
  report<<"    node.style.visibility = \"hidden\";"<<endl;
  report<<"    node.style.height = \"0\";"<<endl;
  report<<"  }"<<endl;
  report<<"}"<<endl;
  report<<"</script>"<<endl;
  report<<"</head>"<<endl;
  report<<"<body>"<<endl;
  report<<endl;
 
  // index title
  report<<"<h1>UPV analysis over "<<target->GetName()<<" </h1>"<<endl;
  report<<"<hr />"<<endl;
  report<<endl;
  
  if(subtitle.compare("")){
    report<<"<pre>"<<subtitle<<"</pre>"<<endl;
    report<<"<hr />"<<endl;
    report<<endl;
  }
  
  // index summary
  report<<"<h2>Summary</h2>"<<endl;
  tm utc;
  time_t timer_stop;
  time ( &timer_stop );
  struct tm * ptm = gmtime ( &timer_stop );
  report<<"<table>"<<endl;
  report<<"  <tr><td>UPV run by:</td><td>"<<getenv("USER")<<"</td></tr>"<<endl;
  report<<"  <tr><td>UPV processing time:</td><td>"<<(int)(timer_stop-timer_start)/3600<<"h, "<<((int)(timer_stop-timer_start)%3600)/60<<"min</td></tr>"<<endl;
  report<<"  <tr><td>Processing Date:</td><td>"<<asctime(ptm)<<" (UTC)</td></tr>"<<endl;
  GPSToUTC (&utc, insegments->GetStart(0));
  report<<"  <tr><td>Requested start:</td><td>"<<(int)insegments->GetStart(0)<<" &rarr; "<<asctime(&utc)<<" (UTC)</td></tr>"<<endl;
  GPSToUTC (&utc, insegments->GetEnd(insegments->GetNsegments()-1));
  report<<"  <tr><td>Requested stop:</td><td>"<<(int)insegments->GetEnd(insegments->GetNsegments()-1)<<" &rarr; "<<asctime(&utc)<<" (UTC)</td></tr>"<<endl;
  report<<"  <tr><td>Requested livetime:</td><td>"<<(int)insegments->GetLiveTime()<<" sec &rarr; "<<setprecision(3)<<fixed<<insegments->GetLiveTime()/3600.0/24<<" days</td></tr>"<<endl;
  insegments->Write(outdir+"/UPV.insegments.txt");
  report<<"  <tr><td>Requested segments:</td><td><a href=\"./UPV.insegments.txt\">UPV.insegments.txt</a></td></tr>"<<endl;
  system(("cp -f "+optionfile+" "+outdir+"/UPV.parameters.txt").c_str());
  report<<"  <tr><td>Number of source channels:</td><td>"<<GetNsource()<<"</td></tr>"<<endl;
  report<<"</table>"<<endl;
  report<<"<hr />"<<endl;
  report<<endl;

  report<<"<h2>Parameters</h2>"<<endl;
  report<<"<table>"<<endl;
  report<<"  <tr><td>Configuration:</td><td><a href=\"./UPV.parameters.txt\">UPV.parameters.txt</a></td></tr>"<<endl;
  double frmin=target->GetFrequencyMin();
  double frmax=target->GetFrequencyMax();
  if(target_freqmin>frmin) frmin=target_freqmin;
  if(target_freqmax<frmax) frmax=target_freqmax;
  report<<"  <tr><td>Target frequency band:</td><td>"<<frmin<<" Hz &lt; f &lt; "<<frmax<<" Hz</td></tr>"<<endl;
  report<<"  <tr><td>Target SNR selection:</td><td>SNR &gt; "<<target_snrmin<<"</td></tr>"<<endl;
  report<<"  <tr><td>Veto definition:</td><td>use-percentage &gt; "<<veto_upmin<<" and number of used source clusters &gt; "<<veto_umin<<"</td></tr>"<<endl;
  report<<"</table>"<<endl;
  report<<"<hr />"<<endl;
  report<<endl;

  report<<"<h2>Target = "<<target->GetName()<<"</h2>"<<endl;
  report<<"<table>"<<endl;
  report<<"  <tr>"<<endl;
  report<<"    <td><a href=\"./"<<target->GetName()<<"/perf/"<<target->GetNameConv()<<"-all_freqtime.png\"><img src=\"./"<<target->GetName()<<"/perf/"<<target->GetNameConv()<<"-all_freqtime.png\" width=\"300\"></td>"<<endl;
  report<<"    <td><a href=\"./"<<target->GetName()<<"/perf/"<<target->GetNameConv()<<"-all_snrtime.png\"><img src=\"./"<<target->GetName()<<"/perf/"<<target->GetNameConv()<<"-all_snrtime.png\" width=\"300\"></td>"<<endl;
  report<<"    <td><a href=\"./"<<target->GetName()<<"/perf/"<<target->GetNameConv()<<"-all_snrfreq.png\"><img src=\"./"<<target->GetName()<<"/perf/"<<target->GetNameConv()<<"-all_snrfreq.png\" width=\"300\"></td>"<<endl;
  report<<"  </tr>"<<endl;
  report<<"</table>"<<endl;
  report<<"<hr />"<<endl;
  report<<endl;


  if(veto_perf[0]==1){
    report<<"<h2>Veto performance</h2>"<<endl;
    report<<"<p>The veto performance has been measured. See the <a href=\"./"<<target->GetName()<<"/perf/index.html\">VetoPerf report</a>.</p>"<<endl;
    report<<"<hr />"<<endl;
  }
  
  // loop over sources
  for(int s=0; s<(int)aIndexes.size(); s++){

    // do not display sources with no clusters
    if(!source_results[aIndexes[s]]->source_nc) continue;
    
    report<<"<h2><a name=\"S"<<s<<"\"></a><font color=\"red\">S"<<s<<":</font> Source = "<<source_results[aIndexes[s]]->source_name<<" ("<<(double)(source_results[aIndexes[s]]->target_veto_u)/(double)(source_results[aIndexes[s]]->target_nc)*100.0<<"%) <a href=\"javascript:void(0)\" name=\""<<source_results[aIndexes[s]]->source_name<<"\" onclick=\"toggle('src_"<<source_results[aIndexes[s]]->source_name<<"')\">[click here to expand/hide]</a> <a href=\"#S"<<s<<"\">[link to here]</a></h2>"<<endl;
    report<<"<div id=\"src_"<<source_results[aIndexes[s]]->source_name<<"\" style=\"visibility:hidden;height:0;\">"<<endl;
    report<<"  <table>"<<endl;
    report<<"    <tr>"<<endl;

    report<<"      <td>"<<endl;
    
    report<<"        <table class=\"list\">"<<endl;

    // target section
    report<<"          <tr><td><b>Target:</b></td><td></td></tr>"<<endl;
    report<<"          <tr><td><img src=\"./qm.gif\" alt=\"number of active target clusters\" title=\"number of active target clusters\"/> &rarr; <a href=\"javascript:showPlot('"<<target->GetName()<<"', '"<<source_results[aIndexes[s]]->source_name<<"', 'target_c_cum');\">number of clusters</a>:</td><td>"<<source_results[aIndexes[s]]->target_nc<<"</td></tr>"<<endl;

    // source section
    report<<"          <tr><td><hr/></td><td></td></tr>"<<endl;
    report<<"          <tr><td><b>Source:</b></td><td></td></tr>"<<endl;
    report<<"          <tr><td><img src=\"./qm.gif\" alt=\"livetime for source triggers\" title=\"livetime for source triggers\"/> &rarr; livetime:</td><td>"<<source_results[aIndexes[s]]->source_livetime<<" s</td></tr>"<<endl;
    report<<"          <tr><td><img src=\"./qm.gif\" alt=\"number of active source clusters\" title=\"number of active source clusters\"/> &rarr; <a href=\"javascript:showPlot('"<<target->GetName()<<"', '"<<source_results[aIndexes[s]]->source_name<<"', 'source_c_cum');\">number of clusters</a>:</td><td>"<<source_results[aIndexes[s]]->source_nc<<"</td></tr>"<<endl;

    // coincidence section
    report<<"          <tr><td><hr/></td><td></td></tr>"<<endl;
    report<<"          <tr><td><b>Coincidence:</b></td><td></td></tr>"<<endl;

    // coinc livetime
    report<<"          <tr><td><img src=\"./qm.gif\" alt=\"livetime of coincident segments (source/target)\" title=\"livetime of coincident segments (source/target)\"/> &rarr; coincident livetime:</td><td>"<<source_results[aIndexes[s]]->coinc_livetime<<" s ("<<source_results[aIndexes[s]]->coinc_livetime/insegments->GetLiveTime()*100.0<<" %)</td></tr>"<<endl;

    // number of coincs (source)
    report<<"          <tr><td><img src=\"./qm.gif\" alt=\"number of source clusters in coincidence with a target cluster\" title=\"number of source clusters in coincidence with a target cluster\"/> &rarr; <a href=\"javascript:showPlot('"<<target->GetName()<<"', '"<<source_results[aIndexes[s]]->source_name<<"', 'source_co_cum');\">number of source clusters</a>:</td><td>"<<source_results[aIndexes[s]]->source_coinc_nc<<" ("<<source_results[aIndexes[s]]->source_coinc_nc/(double)(source_results[aIndexes[s]]->source_nc)*100.00<<" %)</td></tr>"<<endl;

    // number of coincs (target)
    report<<"          <tr><td><img src=\"./qm.gif\" alt=\"number of target clusters in coincidence with a source cluster\" title=\"number of target clusters in coincidence with a source cluster\"/> &rarr; number of target clusters:</td><td>"<<source_results[aIndexes[s]]->target_coinc_nc<<" ("<<source_results[aIndexes[s]]->target_coinc_nc/(double)(source_results[aIndexes[s]]->target_nc)*100.00<<" %)</td></tr>"<<endl;

    // source use-percentage
    report<<"          <tr><td><img src=\"./qm.gif\" alt=\"fraction of source clusters in coincidence with a target cluster\" title=\"fraction of source clusters in coincidence with a target cluster\"/> &rarr; <a href=\"javascript:showPlot('"<<target->GetName()<<"', '"<<source_results[aIndexes[s]]->source_name<<"', 'source_up_cum');\">source use-percentage</a>:</td><td>"<<source_results[aIndexes[s]]->source_coinc_nc<<"/"<<source_results[aIndexes[s]]->source_nc<<" = "<<(double)(source_results[aIndexes[s]]->source_coinc_nc)/(double)(source_results[aIndexes[s]]->source_nc)*100.0<<" %</td></tr>"<<endl;

    // coinc snr vs. snr
    report<<"          <tr><td><img src=\"./qm.gif\" alt=\"target SNR vs. source SNR for coincident clusters\" title=\"target SNR vs. source SNR for coincident clusters\"/> &rarr; <a href=\"javascript:showPlot('"<<target->GetName()<<"', '"<<source_results[aIndexes[s]]->source_name<<"', 'coinc_snrsnr');\">SNR-SNR</a></td><td></td></tr>"<<endl;

    // veto section
    report<<"          <tr><td><hr/></td><td></td></tr>"<<endl;
    report<<"          <tr><td><b>Veto:</b></td><td></td></tr>"<<endl;

    // number of veto segments
    report<<"          <tr><td><img src=\"./qm.gif\" alt=\"number of veto segments\" title=\"number of veto segments\"/> &rarr; number of veto segments:</td><td>"<<source_results[aIndexes[s]]->veto_nseg<<"</td></tr>"<<endl;

    // veto dead-time
    report<<"          <tr><td><img src=\"./qm.gif\" alt=\"total time of source clusters which are above the veto threshold and involved in a coincidence\" title=\"total time of source clusters which are above the veto threshold and involved in a coincidence\"/> &rarr; dead-time:</td><td>"<<source_results[aIndexes[s]]->veto_dt<<" s ("<<source_results[aIndexes[s]]->veto_dt/source_results[aIndexes[s]]->coinc_livetime*100.0<<" %)</td></tr>"<<endl;

    // veto use-percentage
    report<<"          <tr><td><img src=\"./qm.gif\" alt=\"number of source clusters which are above the veto threshold and used to veto target clusters\" title=\"number of source clusters which are above the veto threshold and used to veto target clusters\"/> &rarr; number of used source clusters:</td><td>"<<source_results[aIndexes[s]]->source_veto_u<<" ("<<(double)(source_results[aIndexes[s]]->source_veto_u)/(double)(source_results[aIndexes[s]]->source_veto_nc)*100.0<<" %)</td></tr>"<<endl;

    // veto efficiency
    report<<"          <tr><td><img src=\"./qm.gif\" alt=\"number of target clusters vetoed by a source cluster\" title=\"number of target clusters vetoed by a source cluster\"/> &rarr; number of vetoed target clusters:</td><td>"<<source_results[aIndexes[s]]->target_veto_u<<" ("<<(double)(source_results[aIndexes[s]]->target_veto_u)/(double)(source_results[aIndexes[s]]->target_nc)*100.0<<" %)</td></tr>"<<endl;

    // veto segments
    report<<"          <tr><td><img src=\"./qm.gif\" alt=\"text file with veto segments\" title=\"text file with veto segments\"/> &rarr; veto segments:</td><td><a href=\"./"<<target->GetName()<<"/"<<source_results[aIndexes[s]]->source_name<<"/veto_"<<source_results[aIndexes[s]]->source_name<<".txt\" target=\"_blank\">segment file</a></td></tr>"<<endl;

    // veto recipe
    report<<"          <tr><td><img src=\"./qm.gif\" alt=\"root file with optimized SNR threshold\" title=\"root file with optimized SNR threshold\"/> &rarr; veto recipe:</td><td><a href=\"./"<<target->GetName()<<"/"<<source_results[aIndexes[s]]->source_name<<"/veto_recipe_"<<source_results[aIndexes[s]]->source_name<<".root\" target=\"_blank\">recipe file</a></td></tr>"<<endl;

    report<<"        </table>"<<endl;
    report<<"      </td>"<<endl;


    // plot area
    report<<"      <td>"<<endl;
    report<<"        <img id=\"plot_"<<source_results[aIndexes[s]]->source_name<<"\" src=\"./"<<target->GetName()<<"/"<<source_results[aIndexes[s]]->source_name<<"/source_up_cum.png\" />"<<endl;
    report<<"      </td>"<<endl;

    report<<"    </tr>"<<endl;
    report<<"  </table>"<<endl;
    report<<"</div>"<<endl;
    report<<"<hr/>"<<endl;
    report<<endl;
  }

    
  // index footer
  report<<"<table>"<<endl;
  report<<"  <tr><td>Florent Robinet, <a href=\"mailto:robinet@lal.in2p3.fr\">robinet@lal.in2p3.fr</a></td></tr>"<<endl;
  report<<"</table>"<<endl;
  report<<"</body>"<<endl;
  report<<"</html>"<<endl;
  report.close();

  // veto perf report (if requested)
  if(veto_perf[0]==1){
    C2->SetTriggerSamples(target, target, 0.0, 0.0, insegments,false);// target cluster selection
    string vfiles="";
    for(int s=0; s<(int)aIndexes.size(); s++){
      if(!source_results[aIndexes[s]]->target_veto_u) continue;// remove null vetoes
      vfiles+=source_results[aIndexes[s]]->outdir+"/veto_"+source_results[aIndexes[s]]->source_name+".txt ";
    }

    // combined veto
    if(veto_perf[1]==1){
      Segments *CombinedSeg = new Segments();
      Segments *seg;
      for(int s=0; s<(int)aIndexes.size(); s++){
	if(!source_results[aIndexes[s]]->target_veto_u) continue;// remove null vetoes
	seg = new Segments(source_results[aIndexes[s]]->outdir+"/veto_"+source_results[aIndexes[s]]->source_name+".txt");
	for(int i=0; i<seg->GetNsegments(); i++) CombinedSeg->AddSegment(seg->GetStart(i), seg->GetEnd(i));
	delete seg;
      }
      CombinedSeg->Write(outdir+"/"+target->GetName()+"/perf/veto_combined.txt");
      delete CombinedSeg;
      vfiles+=outdir+"/"+target->GetName()+"/perf/veto_combined.txt";
    }

    // make performance
    int nveto=vp->SetVetoes(vfiles);
    for(int v=0; v<nveto; v++) vp->MakeVetoPerformance(v);
    vp->MakeReport();
  }

  return true;
}

////////////////////////////////////////////////////////////////////////////////////
vector <int> UPV::Sort(void){
////////////////////////////////////////////////////////////////////////////////////
  vector <int> s_index;// sorted indexes

  if(!GetNsource()) return s_index;

  vector <double> s_sig;
  bool sorted;

  // init
  s_index.push_back(0);
  if(source_results[0]->target_nc) s_sig.push_back(GetVetoEfficiency(0));
  else s_sig.push_back(0.0); // no target cluster time --> 0 efficiency
  
  for(int s=1; s<GetNsource(); s++){
   
    // no target cluster time --> 0 efficiency
    if(!source_results[s]->target_nc){
      s_index.push_back(s);
      s_sig.push_back(0.0);
      continue;
    }
      
    sorted=false;
    for(int i=0; i<(int)s_index.size(); i++){
      if(GetVetoEfficiency(s)>=s_sig[i]){
	s_index.insert(s_index.begin()+i,s);
	s_sig.insert(s_sig.begin()+i,GetVetoEfficiency(s));
	sorted=true;
	break;
      }
    }
    if(!sorted){
      s_index.push_back(s);
      s_sig.push_back(GetVetoEfficiency(s));
    }
  }
  
  s_sig.clear();

  return s_index;
}

////////////////////////////////////////////////////////////////////////////////////
int UPV::GetWinIndex(void){
////////////////////////////////////////////////////////////////////////////////////
  int windex = -1;
  double val, bestval=-1;

  // find best source efficiency
  for(int s=0; s<GetNsource(); s++){
    if(!GetSourceStatus(s)) continue;
    if(source_results[s]->target_nc) val=GetVetoEfficiency(s);// veto efficiency
    else val=0.0;
    if(val>bestval){
      bestval=val;
      windex=s;
    }
  }
  
  return windex;
}
