//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#include "UPV.h"

int printhelp(void){
  cerr<<endl;
  cerr<<"upv:"<<endl;
  cerr<<"This program runs a Glitch-To-Glitch analysis."<<endl;
  cerr<<endl;
  cerr<<"Usage:"<<endl; 
  cerr<<endl;
  cerr<<"CASE1: upv [GPS start time] [GPS stop time] [option file]"<<endl; 
  cerr<<"|__ runs the upv algorithm between 2 GPS times (integers)."<<endl; 
  cerr<<endl;
  cerr<<"CASE2: upv [segment file] [option file]"<<endl; 
  cerr<<"|__ runs the upv algorithm over a segment list."<<endl; 
  cerr<<endl;
  cerr<<">>> In all cases, the user must must provide an option file listing"<<endl;
  cerr<<"    the upv parameters."<<endl;
  cerr<<">>> In all cases, the GPS times must be integer values"<<endl;
  cerr<<endl;
  cerr<<"-------------------------------------------------------------------"<<endl;
  cerr<<endl;
  cerr<<"Optionally, one last argument, \"hierarchy\" can be given,"<<endl;
  cerr<<"upv is then run in a hierarchical mode."<<endl;
  cerr<<endl;
  return 1;
}

int main (int argc, char* argv[]){


  //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  //--------------              COMMAND LINE              --------------
  //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  if(argc>1&&!((string)argv[1]).compare("version")){
    PrintVersionUPV();
    return 0;
  }

  int nargs=argc;
  
  // hierarchy
  bool hierarchy=false;
  if(argc&&!strcmp(argv[argc-1],"hierarchy")){
    hierarchy=true;
    nargs=argc-1;
  }
  
  // number of arguments
  if(nargs!=4&&nargs!=3) return printhelp();

  int start;
  int stop;
  Segments *insegments;
  string optionfile;
  string segmentfile;

  //**** CASE1: start and stop time
  if(nargs==4){
    start=(int)round(atof(argv[1]));
    stop=(int)round(atof(argv[2]));
    optionfile=(string)argv[3];
 
    // segment object
    insegments = new Segments(start,stop);
  }

  //****** CASE2: segment file
  else if(atoi(argv[1])<700000000&&IsTextFile(argv[1])){
    segmentfile=(string)argv[1];
    optionfile=(string)argv[2];

    // load segment file
    insegments = new Segments(segmentfile);
  }

  //******
  else{
    cerr<<"upv: the option sequence cannot be interpreted."<<endl;
    cerr<<"Type 'upv' for help."<<endl;
    return -1;
  }

  // check timing
  if(!insegments->GetLiveTime()){
    cerr<<"upv: the input segments are corrupted."<<endl;
    cerr<<"Type 'upv' to get help"<<endl;
    delete insegments;
    return -1;
  }
  
  // analysis time period
  start=insegments->GetStart(0);
  stop=insegments->GetEnd(insegments->GetNsegments()-1);

  // check that the option file exists
  if(!IsTextFile(optionfile)){
    cerr<<"upv: the option file "<<optionfile<<" cannot be found."<<endl;
    cerr<<"Type 'upv' for help."<<endl;
    delete insegments;
    return -1;
  }
 
  // init UPV object
  UPV *U = new UPV(optionfile,insegments);
  vector <int> sorted_indexes;

  // command
  string scommand="";
  for(int a=0; a<argc; a++) scommand+=(string)argv[a]+" ";
  U->SetSubTitle(scommand);

  // **************** WITHOUT HIERARCHY ****************
  if(!hierarchy){

    // loop over source triggers
    for(int i=0; i<U->GetNsource(); i++){
      
      // load source triggers
      if(!U->LoadSourceTriggers(i)) continue;
      cout<<U->GetTargetName()<<" <--- "<<U->GetSourceName()<<endl;
      
      // perform coincidence
      if(U->MakeCoinc()<=0) continue;
      
      // make monitors
      if(!U->MakeTriggerMonitors(0)) continue;
      if(!U->MakeTriggerMonitors(1)) continue;
      if(!U->MakeCoincMonitors()) continue;
      U->MakeProcMonitors();

      // make veto
      U->MakeVetoThreshold();
      cout<<"\tveto efficiency = "<<U->GetVetoEfficiency(i)<<endl;

      // exclude zero-efficiency sources
      if(!U->GetVetoEfficiency(i)) continue;
      
      // print output
      U->PrintMonitors();
      U->PrintVeto();
    }

    // sort results by efficiency
    sorted_indexes=U->Sort();

  }

  // **************** WITH HIERARCHY ****************
  else{
    
    int winindex;
    
    // rounds
    for(int r=0; r<U->GetNsource(); r++){
      cout<<endl;
      cout<<"****************************************************"<<endl;
      cout<<"***************    ROUND "<<r<<"    ***************"<<endl;
      cout<<"****************************************************"<<endl;

      // loop over source triggers
      for(int i=0; i<U->GetNsource(); i++){

	// skip if channels is de-activated
	if(!U->GetSourceStatus(i)) continue;
	
	// load source triggers
	if(!U->LoadSourceTriggers(i)){ U->SetSourceStatus(i, false); continue; }
	cout<<U->GetTargetName()<<" <--- "<<U->GetSourceName()<<endl;
	
	// perform coincidence
	if(U->MakeCoinc()<=0){ U->SetSourceStatus(i, false); continue; }
	
	// make monitors
	if(!U->MakeTriggerMonitors(0)){ U->SetSourceStatus(i, false); continue; }
	if(!U->MakeTriggerMonitors(1)){ U->SetSourceStatus(i, false); continue; }
	if(!U->MakeCoincMonitors()){ U->SetSourceStatus(i, false); continue; }
	U->MakeProcMonitors();

	// make veto
	U->MakeVetoThreshold();
	cout<<"\tveto efficiency = "<<U->GetVetoEfficiency(i)<<endl;

	// exclude zero-efficiency sources
	if(!U->GetVetoEfficiency(i)) U->SetSourceStatus(i, false);
      }

      // get winning channel
      winindex=U->GetWinIndex();
      if(winindex<0) break;// no channel left

      // stop here if veto efficiency is 0
      if(!U->GetVetoEfficiency(winindex)) break;

      // save index
      sorted_indexes.push_back(winindex);
   
      // make plots and print results
      U->LoadSourceTriggers(winindex);
      cout<<"\n>>>>>> winning channel = "<<U->GetSourceName()<<endl;
      U->MakeCoinc();
      U->MakeTriggerMonitors(0);
      U->MakeTriggerMonitors(1);
      U->MakeCoincMonitors();
      U->MakeProcMonitors();
      U->MakeVetoThreshold();
      U->PrintMonitors();
      U->PrintVeto();

      // exclude channels for next rounds
      U->SetSourceStatus(winindex, false);

      // print status
      cout<<"number of remaining source channels after this round = "<<U->GetSourceNactive()<<endl<<endl;
      
      // remove veto triggers
      if(!U->RemoveVetoedTriggers()) break;
    }

  }
  
  // generate HTML report
  U->MakeHtml(sorted_indexes);
  sorted_indexes.clear();


  // VetoPerf analysis
  
  delete U;
  return 0;
}

